# Autogenerated by Astropy's setup.py on 2013-11-20 20:09:10.840506

from astropy.version_helpers import update_git_devstr, get_git_devstr

_last_generated_version = '0.3'

version = update_git_devstr(_last_generated_version)
githash = get_git_devstr(sha=True, show_warning=False)

major = 0
minor = 3
bugfix = 0

release = True
debug = False

try:
    from .utils._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
