# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This packages contains python packages that are bundled with Astropy
but are external to Astropy, and hence are developed in a separate
source tree.  Note that this package is distinct from the /cextern
directory of the source code distribution, as that directory only
contains C extension code.
"""
