"""
This package contains the sphinx extensions used by Astropy.

The `automodsumm` and `automodapi` modules contain extensions used by Astropy
to generate API documentation - see the docstrings for details.  The `numpydoc`
module is dervied from the documentation tools numpy and scipy use to parse
docstrings in the numpy/scipy format, and are also used by Astropy.  The
other modules are dependencies for `numpydoc`.
"""
